from typing import Any, Callable, Deque, Generic, Optional, TypeVar
from threading import Lock, Thread
from collections import deque
from timeit import default_timer as timer

import numpy as np

import logging
import time
import copy

ContentType = TypeVar('ContentType')

logger = logging.getLogger(__name__)


class AsyncDispatch(Generic[ContentType]):
    def __init__(self, target: Callable[[ContentType], bool], running_condition: Callable[[], bool], *,
                 copy_fun: Callable[[Any], Any] = copy.copy, maxlen: int = 1,
                 finish_batch_when_terminated: bool = False) -> None:
        self.__lock: Lock = Lock()

        self.__target: Callable[[ContentType.__args__], bool] = target
        self.__running_condition: Callable[[], bool] = running_condition
        self.__copy_fun: Callable[[Any], Any] = copy_fun
        self.__finish_batch_when_terminated: bool = finish_batch_when_terminated

        self.__payload: Deque[ContentType] = deque(maxlen=maxlen)
        self.__running: bool = False
        self.__terminate_after_last: bool = False
        self.__thread: Thread = Thread(target=self.run)

    def pop(self) -> Optional[ContentType]:
        with self.__lock:
            try:
                return self.__copy_fun(self.__payload.pop())
            except IndexError:
                return None

    def update(self, new_payload: ContentType) -> None:
        with self.__lock:
            self.__payload.append(self.__copy_fun(new_payload))

    def __enter__(self) -> Optional[ContentType]:
        self.__lock.acquire()
        return self.__payload

    def __exit__(self, ) -> None:
        self.__lock.release()

    def join_if_alive(self, timeout=None):
        self.__running = False
        return not self.__thread.is_alive() or self.__thread.join(timeout)

    def join_after_last(self):
        self.__terminate_after_last = True
        return not self.__thread.is_alive() or self.__thread.join()

    def start(self):
        self.__thread.start()
        return self

    def is_running(self):
        return self.__running and (self.__finish_batch_when_terminated or self.__running_condition())

    def is_terminated_after_last(self):
        return self.__terminate_after_last or (self.__finish_batch_when_terminated and not self.__running_condition())

    def run(self):
        self.__running: bool = True
        self.__terminate_after_last: bool = False

        try:
            while self.is_running():
                current_item: Optional[ContentType] = self.pop()

                if current_item is None:
                    if self.is_terminated_after_last():
                        self.__running = False
                    else:
                        time.sleep(0.5)
                    continue

                exit_code = self.__target(*current_item)

                if exit_code is not None and not exit_code:
                    self.update(current_item)
                    time.sleep(2.0)

        except Exception as e:
            function_name: Optional[str] = getattr(self.__target, '__name__', None)
            if function_name is None:
                function_name = getattr(self.__target, 'args')
                if len(function_name):
                    function_name = getattr(function_name[-1], '__name__', '?')

            logger.error(f'An error occurred during "{function_name}": {e}.')
            raise e from None


def measure_time(*arg, num=100, log_function=logger.info) -> Callable:
    """
    Measures the execution of an annotated function if in debug mode.
    """

    def __func__(fun):
        fun._i = -1
        fun._f = np.empty(num)

        def __inner__(*args, **kwargs):
            fun._i += 1
            ii = fun._i % fun._f.shape[0]

            fun._f[ii] = timer()
            l = fun(*args, **kwargs)
            fun._f[ii] = timer() - fun._f[ii]

            if not ii and fun._i:
                minimum, average, maximum = np.min(fun._f), np.mean(fun._f), np.max(fun._f)
                log_function(f'{fun.__name__}: min:{minimum:0.2f}, avg:{average:0.2f}, max:{maximum:0.2f} [s].')
            return l

        return __inner__ if __debug__ else fun

    if len(arg) == 1:
        if type(arg[0]) == int:
            num = arg[0]
        else:
            return __func__(*arg)

    return __func__
